
from sqlalchemy import Column, Integer, String, Sequence, text
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base

from langchain.cache import SQLAlchemyCache
import langchain

Base = declarative_base()


class FulltextLLMCache(Base):  # type: ignore
    """Table for indexed LLM Cache"""

    __tablename__ = "llm_cache_fulltext"

    id = Column(Integer, Sequence('cache_id'), primary_key=True)
    prompt = Column(String, nullable=False, index=True)
    llm = Column(String, nullable=False)

    idx = Column(Integer)

    response = Column(String)



def configure_caching(database_uri):
    engine = create_engine(database_uri)
    langchain.llm_cache = SQLAlchemyCache(engine, FulltextLLMCache)



import time
from typing import Any

from langchain.tools import BaseTool
from sqlalchemy import text

from qabot.duckdb_query import run_sql_catch_error


class DuckDBTool(BaseTool):
    name = "execute"
    description = """useful for when you need to run SQL queries against a DuckDB database."""

    database: Any = None

    def __init__(self, engine, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.database = engine

    def _run(self, query: str) -> str:

        #time.sleep(2)
        #print(query)
        return run_sql_catch_error(self.database, query)


    async def _arun(self, query: str) -> str:
        raise NotImplementedError("DuckDBTool does not support async")





import textwrap

from langchain import LLMMathChain
from langchain.agents import Tool, initialize_agent
from langchain.llms import OpenAIChat

from qabot.data_loader_chain import get_duckdb_data_loader_chain
from qabot.data_query_chain import get_duckdb_data_query_chain
from qabot.duckdb_query import run_sql_catch_error
from qabot.tools.describe_duckdb_table import describe_table_or_view


#from qabot.duckdb_documentation import get_duckdb_docs_chain


def create_agent_executor(
        database_engine=None,
        tables=None,
        return_intermediate_steps=False,
        callback_manager=None,
        verbose=False,
):

    # db_kwargs = dict(
    #     include_tables=tables,
    #     sample_rows_in_table_info=3
    # )
    # if database_engine is not None:
    #     db = SQLDatabase(database_engine, **db_kwargs)
    # elif database_uri is not None:
    #     db = SQLDatabase.from_uri(
    #         database_uri,
    #         **db_kwargs
    #     )
    # else:
    #     raise ValueError("Must provide either database_uri or database_engine")


    llm = OpenAIChat(
        model_name="gpt-3.5-turbo",
        temperature=0.0
    )


    calculator_chain = LLMMathChain(llm=llm, verbose=False)
    # db_chain = SQLDatabaseChain(llm=llm, database=db,
    #                             return_intermediate_steps=return_intermediate_steps,
    #                             verbose=False)

    db_chain = get_duckdb_data_query_chain(llm=llm, database=database_engine)


    data_loader_chain = get_duckdb_data_loader_chain(llm=llm, database=database_engine)
    #duckdb_docs_qa_chain = get_duckdb_docs_chain(llm=llm)

    tools = [
        #DuckDBTool(db),
        Tool(
            name="Calculator",
            func=calculator_chain.run,
            description="Useful for when you need to answer questions about math"
        ),
        # Tool(
        #     name="DuckDB QA System",
        #     func=duckdb_docs_qa_chain.run,
        #     description="useful for when you need to answer questions about duckdb. Input should be a fully formed question."
        # ),
        Tool(
            name="Show Tables",
            func=lambda _: run_sql_catch_error(database_engine, "show tables;"),
            description="Useful to show the available tables and views. Empty input required."
        ),
        Tool(
            name="Describe Table",
            func=lambda table: describe_table_or_view(database_engine, table),
            description="Useful to show the column names and types of a table or view. Use the table name as the input."
        ),
        Tool(
            name="Data Op",
            func=lambda input: db_chain({
                'table_names': run_sql_catch_error(database_engine, "show tables;"),
                'input': input}),
            description=textwrap.dedent("""useful for when you need to operate on data and answer questions
            requiring data. Input should be in the form of a natural language question containing full context 
            including what tables and columns are relevant to the question. Use only after data is present and loaded.
            """,)
        ),
        # Tool(
        #     name="Data Loader",
        #     func=data_loader_chain.run,
        #     description="""useful for when you need to load data from a specific local or remote file.
        #     Files can be a csv, json, sqlite, or parquet.
        #     """
        # )
    ]


    agent = initialize_agent(
        tools,
        llm,
        #agent="conversational-react-description",
        agent="zero-shot-react-description",
        callback_manager=callback_manager,
        return_intermediate_steps=return_intermediate_steps,
        verbose=True,
        agent_kwargs={
            "prefix": prompt_prefix_template.format(table_names=run_sql_catch_error(database_engine, "show tables;"))
        }
    )
    #agent.agent.llm_chain.prompt.template
    return agent


prompt_prefix_template = """Answer the following question as best you can by querying for data to back up
your answer. 

Refuse to delete any data, or drop tables. When answering, you MUST query the database for any data. 
Check the available tables exist first. Prefer to create views of data, then select data from the view.

It is important that you use the exact phrase "Final Answer:" in your final answer.
You always provide the SQL queries you ran as part of your final answer.

You have access to the following data tables:
{table_names}

You have access to the following tools:
"""


from langchain import LLMChain, PromptTemplate
from langchain.agents import AgentExecutor, ZeroShotAgent

from qabot.duckdb_execute_tool import DuckDBTool


def get_duckdb_data_loader_chain(llm, database):
    tool = DuckDBTool(engine=database)

    prompt = PromptTemplate(
        input_variables=["input", "agent_scratchpad"],
        template=_DEFAULT_TEMPLATE,
    )

    llm_chain = LLMChain(llm=llm, prompt=prompt)
    tool_names = [tool.name]
    agent = ZeroShotAgent(llm_chain=llm_chain, allowed_tools=tool_names)
    agent_executor = AgentExecutor.from_agent_and_tools(
        agent=agent,
        tools=[tool],
        verbose=True
    )

    return agent_executor


_DEFAULT_TEMPLATE = """Given a description of data containing a url or local path, identify the 
extension (.json, .parquet, .csv), generate an appropriate table name, generate the SQL to load the
data into a DuckDB database.

CREATE TABLE test AS SELECT * FROM 'input-file';

Use the following format:

Input: "Description of input data here"
Extension: Extension of the input data
Action: the action to take, should be one of [execute]
Action Input: "SQL Query to load data"
Final Answer: Name of table or view that was created. Along with the SQL query that was executed.

For example:

``` 
Input: "open test.parquet from the data folder"
Extension: parquet
Action: execute
Action Input: "CREATE TABLE test AS SELECT * FROM 'data/test.parquet';"
Final Answer: test

The following SQL queries were executed:
CREATE TABLE test AS SELECT * FROM 'data/test.parquet';
```

Another example:
 
```
Input: "import data/records.json as customers"
Extension: json
Action: execute
Action Input: "CREATE table customers AS SELECT * FROM 'data/records.json';"
Final Answer: customers

The following SQL queries were executed:
CREATE table customers AS SELECT * FROM 'data/records.json';
```


Another Example:

Input: "create a view of the covid data from "s3://covid19-lake/data.csv"
Extension: csv
Action: execute
Action Input: "CREATE VIEW covid AS SELECT * FROM 's3://covid19-lake/data.csv';"
Final Answer: covid


In the case of a query that returns no results, you should output a summary as your final answer:
```
Final Answer: The data has been written to a parquet file at 'data/titanic_gender_survival.parquet'
The following SQL queries were executed:
- COPY (select * from titanic) TO 'data/titanic_gender_survival.parquet' (FORMAT PARQUET);
```

Input: {input}
{agent_scratchpad}"""



from typing import List, Optional

from pydantic import BaseSettings

class Settings(BaseSettings):
    OPENAI_API_KEY: str

    QABOT_DATABASE_URI: Optional[str] = None
    QABOT_CACHE_DATABASE_URI = "duckdb:///:memory:"

    QABOT_TABLES: Optional[List[str]]



from langchain import LLMChain, PromptTemplate
from langchain.agents import AgentExecutor, Tool, ZeroShotAgent

from qabot.duckdb_execute_tool import DuckDBTool
from qabot.duckdb_query import run_sql_catch_error
from qabot.tools.describe_duckdb_table import describe_table_or_view


def get_duckdb_data_query_chain(llm, database):
    tools = [
        Tool(
            name="Show Tables",
            func=lambda _: run_sql_catch_error(database, "show tables;"),
            description="Useful to show the available tables and views. Empty input required."
        ),
        Tool(
            name="Describe Table",
            func=lambda table: describe_table_or_view(database, table),
            description="Useful to show the column names and types of a table or view. Use a valid table name as the input."
        ),
        Tool(
            name="Query Inspector",
            func=lambda input: input,
            description="Useful to show the query before execution. Always inspect your query before execution."
        ),
        DuckDBTool(engine=database),
    ]

    # prompt = PromptTemplate(
    #     input_variables=["input", "agent_scratchpad"],
    #     template=_DEFAULT_TEMPLATE,
    # )

    prompt = ZeroShotAgent.create_prompt(
        tools,
        prefix=prefix,
        #suffix=suffix,
        input_variables=["input", "agent_scratchpad", 'table_names'],
    )

    llm_chain = LLMChain(llm=llm, prompt=prompt)

    tool_names = [tool.name for tool in tools]

    agent = ZeroShotAgent(llm_chain=llm_chain, allowed_tools=tool_names)
    agent_executor = AgentExecutor.from_agent_and_tools(
        agent=agent,
        tools=tools,
        verbose=True
    )

    return agent_executor


prefix = """Given an input question, identify the relevant tables and relevant columns, then create
one single syntactically correct DuckDB query to inspect, then execute, before returning the answer. 
If the input is a valid looking SQL query selecting data or creating a view, execute it directly. 

Before answering, you MUST query the database for any data. Inspect your query before execution.

Refuse to delete any data, or drop tables.
 
Unless the user specifies in their question a specific number of examples to obtain, you always limit
your query to at most 5 results. You can order the results by a relevant column to return the most interesting 
examples in the database.

Pay attention to use only the column names that you can see in the schema description. Be careful 
to not query for columns that do not exist. Also, pay attention to which column is in which table.

You always summarize the relevant SQL queries you ran as part of your final answer. An example final 
answer:
```
Final Answer: There were 109 male passengers who survived.
The following SQL queries were executed to obtain the result:
- SELECT Sex, Survived FROM titanic limit 5;
- CREATE VIEW male_survivors AS SELECT * FROM titanic WHERE Sex = 'male' AND Survived = 1;
- select count(*) from male_survivors;
```

In the case of a query that fails or returns no results, you should output a summary as your final answer:
```
Final Answer: The data has been written to a parquet file at 'data/titanic_gender_survival.parquet'
The following SQL queries were executed to obtain the result:
- COPY (select * from titanic) TO 'data/titanic_gender_survival.parquet' (FORMAT PARQUET);
```


It is important that you use the exact phrase "Final Answer:" in your final answer.

Queries should be output across multiple lines for readability and don't use any escape characters. 
You have access to the following tables/views:
{table_names}

You have access to the following tools:
"""



# Other examples

"""Examples:?

For example:
 
Input: "Create a names table with an id, name and email column"
Thought: "I need to execute a query to create a table called names, with an id, name and email column"
Action: execute
Action Input: "CREATE TABLE names (id INTEGER, name VARCHAR, email VARCHAR);"
Thought: "I should describe the table to make sure it was created correctly"
Action: Describe Table
Action Input: names
Final Answer: 


Errors should be returned directly:

Input: "Create a names table with an id, name and email column"
Thought: "I need to execute a query to create a table called names, with an id, name and email column"
Action: execute
Action Input: "CREATE TABLE names (id INTEGER, name VARCHAR, email VARCHAR);"
Final Answer: Error: Catalog Error: Table with name "names" already exists!


For example:
 
Input: "count the number of entries in the "addresses" table that belong to each different city filtering out cities with a count below 50"
Thought: "I need to execute a query to count the number of entries in the "addresses" table that belong to each different city filtering out cities with a count below 50"
Action: execute
Action Input: SELECT city, COUNT(*) FROM addresses GROUP BY city HAVING COUNT(*) >= 50 limit 2;
Thought: 
Final Answer:

"""


import textwrap
import time
from typing import List, Optional
import warnings

import langchain
import typer
from langchain.callbacks import OpenAICallbackHandler
from langchain.callbacks.base import CallbackManager
from langchain.schema import AgentAction
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.prompt import Confirm, Prompt
from rich.traceback import install

from qabot.caching import configure_caching
from qabot.config import Settings
from qabot.duckdb_manual_data_loader import create_duckdb_from_files
from qabot.agent import create_agent_executor

install(suppress=[typer, langchain], max_frames=5, show_locals=False)
warnings.filterwarnings("ignore")

INITIAL_NON_INTERACTIVE_PROMPT = "🚀 How can I help you explore your database?"
INITIAL_INTERACTIVE_PROMPT = "[bold green] 🚀 How can I help you explore your database?"
FOLLOW_UP_PROMPT = "[bold green] 🚀 any further questions?"
PROMPT = "[bold green] 🚀 Query"

def format_intermediate_steps(intermediate_steps):
    if isinstance(intermediate_steps, list):
        return "\n".join(intermediate_steps)
    else:
        return str(intermediate_steps)


def format_agent_action(agent_action: AgentAction, observation) -> str:
    result = ''
    if isinstance(observation, dict):
        if 'result' in observation:
            result = observation['result']
        elif 'output' in observation:
            result = observation['output']
        if 'intermediate_steps' in observation:
            observation = format_intermediate_steps(observation['intermediate_steps'])

    return f"""
[red]{agent_action.tool.strip()}[/red]
  [green]{agent_action.tool_input.strip()}[/green]

  [blue]{str(observation).strip()}[/blue]

[bold red]{result}[/bold red]
"""


class QACallback(OpenAICallbackHandler):
    def __init__(self, *args, **kwargs):
        self.progress: Progress = kwargs.pop('progress')
        self.chain_task_ids = []
        self.tool_task_id = None

        super().__init__(*args, **kwargs)

    def on_chain_start(self, serialized, inputs, **kwargs):
        self.chain_task_ids.append(self.progress.add_task(f"on chain start"))
        if isinstance(serialized, dict) and 'name' in serialized:
            self.progress.update(self.chain_task_ids[-1], description=f"[yellow]{serialized['name']}")
        elif 'agent_scratchpad' in inputs and len(inputs['agent_scratchpad']):
            self.progress.update(self.chain_task_ids[-1], description=inputs['agent_scratchpad'])

    def on_agent_action(
        self, action: AgentAction, color: Optional[str] = None, **kwargs
    ):
        """Run on agent action."""
        print(action.log)

    def on_chain_end(self, outputs, **kwargs):
        super().on_chain_end(outputs, **kwargs)
        if isinstance(outputs, dict) and 'output' in outputs:
            outputs = outputs['output']

        self.progress.update(self.chain_task_ids[-1], description=f"[yellow]{outputs}")
        self.progress.remove_task(self.chain_task_ids.pop())

    def on_llm_end(self, response, **kwargs):
        print("[yellow]On llm end")

    def on_tool_start(self, **kwargs):
        print("[yellow]On tool start")

    def on_tool_end(self, output: str, **kwargs):
        print("[yellow]On tool end")

def main(
        query: str = typer.Option("Describe the tables", '-q', '--query', prompt=INITIAL_NON_INTERACTIVE_PROMPT),
        file: Optional[List[str]] = typer.Option(None, "-f", "--file", help="File or url containing data to query"),
        #database_uri: Optional[str] = typer.Option(None, "-d", "--database", help="Database URI (e.g. sqlite:///mydb.db)"),
        table: Optional[List[str]] = typer.Option(None, "--table", "-t", help="Limit queries to these tables (can be specified multiple times)"),
        disable_cache: bool = typer.Option(False, "--disable-cache", help="Disable caching of LLM queries"),
        verbose: bool = typer.Option(False, "-v", "--verbose"),

):
    """
    Query a database using a simple english query.

    Example:
        qabot -q "What is the average age of the people in the table?"
    """

    settings = Settings()
    executed_sql = ''
    # If files are given load data into local DuckDB
    database_engine = None
    if len(file) > 0:
        if isinstance(file, str):
            file = [file]
        print("[red]🦆[/red] [bold]Loading data from files...[/bold]")
        database_engine, executed_sql = create_duckdb_from_files(file)
        executed_sql = '\n'.join(executed_sql)
    else:
        print("[red]🦆[/red]")

    with Progress(
        SpinnerColumn(),
        TextColumn("[green][progress.description]{task.description}"),
        transient=False,
    ) as progress:

        callback_manager = CallbackManager(handlers=[QACallback(progress=progress)])

        if not disable_cache:
            t = progress.add_task(description="Setting up cache...", total=None)
            configure_caching(settings.QABOT_CACHE_DATABASE_URI)
            progress.remove_task(t)

        t2 = progress.add_task(description="Creating LLM agent using langchain...", total=None)

        agent = create_agent_executor(
            #database_uri=database_uri or settings.QABOT_DATABASE_URI,
            database_engine=database_engine,
            tables=table,
            return_intermediate_steps=True,
            callback_manager=callback_manager,
            verbose=False,
        )
        progress.remove_task(t2)
        chat_history = [f"""
        Startup SQL Queries:
        ```
        {executed_sql}
        ```
        """]

        while True:

            t = progress.add_task(description="Processing query...", total=None)
            print("[bold red]Query: [/][green]" + query)

            inputs = {
                "input": chat_history[0] + query,
                #'chat_history': '\n\n'.join(chat_history)
            }

            result = agent(inputs)

            progress.remove_task(t)

            # Show intermediate steps
            if verbose:
                progress.console.print("[bold red]Intermediate Steps: [/]")
                for i, (agent_action, action_input) in enumerate(result['intermediate_steps'], 1):
                    print(f"  [bold red]Step {i}[/]")
                    print(textwrap.indent(format_agent_action(agent_action, action_input), "    "))

                print()

            # Stop the progress before outputting result and prompting for input
            progress.stop()
            print()

            print("[bold red]Result:[/]\n[bold blue]" + result['output'] + "\n")
            chat_history.append(result['output'])

            if not Confirm.ask(FOLLOW_UP_PROMPT, default=True):
                break

            print()
            query = Prompt.ask(PROMPT)

            if query == "exit" and Confirm.ask("Are you sure you want to Quit?"):
                break

            progress.start()



def run():
    typer.run(main)


if __name__ == '__main__':
    run()